package markov;

import java.io.Serializable;
import java.util.Arrays;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;
import java.util.stream.Collectors;

public class Lookup implements Serializable {

    private final LinkedHashMap<Token, Integer> tokens = new LinkedHashMap<>();
    private Decission[] finalData;
    private int[] finishedSums;
    private boolean isFinishedCollecting = false;

    private int totalCounts = 0;

    public int getDistinctTokens() {
        finishGuard();
        return finalData.length;
    }

    private void finishGuard() {
        if (!isFinishedCollecting) {
            throw new IllegalStateException("not jet finished");
        }
    }

    public void add(Token t) {
        if (isFinishedCollecting) {
            throw new IllegalStateException("already finished");
        }
        Integer i = tokens.getOrDefault(t, 0);
        tokens.put(t, i + 1);
        totalCounts++;
    }

    public Decission forRandom(double random) {
        finishGuard();
        if (random < 0 || random >= 1) {
            throw new IllegalArgumentException("expected double [0; 1)");
        }
        int id = (int) Math.floor(random * totalCounts);
        int i = Arrays.binarySearch(finishedSums, id);
        i = Math.min(i >= 0 ? i + 1 : -1 * (i + 1), finishedSums.length - 1);
        return finalData[i];
    }

    public Decission forId(int id) {
        finishGuard();
        if (id < 0 || id > finishedSums.length) {
            throw new IllegalArgumentException(String.format("got %d but expected id [0; %d)",
                    id,
                    finishedSums.length));
        }
        return finalData[id];
    }

    public Decission average() {
        finishGuard();
        return finalData[this.finishedSums.length - 1];
    }

    public Set<Token> allPossible() {
        finishGuard();
        return Arrays.asList(finalData)
                .stream()
                .map(Decission::getToken)
                .collect(Collectors.toSet());
    }

    void finishCollection() {
        if (!isFinishedCollecting) {
            int size = tokens.size();
            finishedSums = new int[size];
            finalData = new Decission[size];
            List<Entry<Token, Integer>> orderedEntries = tokens.entrySet()
                    .stream()
                    .sequential()
                    .sorted(Comparator.comparingInt(Entry::getValue))
                    .collect(Collectors.toList());
            int sum = 0;
            for (int i = 0; i < orderedEntries.size(); i++) {
                Entry<Token, Integer> entry = orderedEntries.get(i);
                sum += entry.getValue();
                finishedSums[i] = sum;
                finalData[i] = new Decission(entry.getKey(),
                        i,
                        this, (double) entry.getValue() / getTotalCounts());
            }
            this.tokens.clear();
            isFinishedCollecting = true;
        }
    }

    int getTotalCounts() {
        return totalCounts;
    }

    void resetFinishding() {
        int sum = 0;
        for (int i = 0; i < finalData.length; i++) {
            this.tokens.put(finalData[i].getToken(), finishedSums[i] - sum);
            sum += finishedSums[i];
        }
        this.isFinishedCollecting = false;
        this.finishedSums = null;
        this.finalData = null;
    }

    @Override
    public String toString() {
        return isFinishedCollecting ? "Lookup" :
               ("Lookup [tokens= " + tokens.entrySet()
                       .stream()
                       .sorted((e1, e2) -> Integer.compare(e2.getValue(),
                               e1.getValue()))
                       .map(e -> String.format("%d*%s", e.getValue(), e.getKey()))
                       .collect(Collectors.joining(", ")) + "]");
    }


}
